import unittest
import pyterrier as pt
if not pt.started():
    pt.init()
from pyterrier_adaptive import CorpusGraph
from pyterrier_pisa import PisaIndex
import tempfile

class TestCorpusGraph(unittest.TestCase):
    def test_vaswani_from_retriever(self):
        dataset = pt.get_dataset('irds:vaswani')
        with tempfile.TemporaryDirectory() as dout:
            indexer = PisaIndex(f'{dout}/index.pisa')
            indexer.index(dataset.get_corpus_iter())
            bm25 = PisaIndex(f'{dout}/index.pisa').bm25(num_results=5)
            graph = CorpusGraph.from_retriever(bm25, dataset.get_corpus_iter(), f'{dout}/graph', k=4)

if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
from pyterrier_adaptive import Laff


class TestLaff(unittest.TestCase):
    def test_laff(self):
        laff = Laff()
        scores = laff.compute_affinity(["hello", "world"], ["world", "hello"])
        self.assertEqual(len(scores), 2)
        self.assertTrue(isinstance(scores[0], float))
        self.assertTrue(isinstance(scores[1], float))

        inp = pd.DataFrame({
            "text": ["hello", "world"],
            "other_text": ["world", "hello"]
        })
        out = laff(inp)
        self.assertEqual(len(out), 2)
        self.assertTrue('affinity' in out.columns)


if __name__ == '__main__':
    unittest.main()



import pickle
import tempfile
from lz4.frame import LZ4FrameFile
import json
import numpy as np
import pandas as pd
from pathlib import Path
import more_itertools
from typing import Union, Tuple, List
import ir_datasets
from npids import Lookup
import pyterrier_alpha as pta

try:
  from functools import cached_property
except ImportError:
  # polyfill for python<3.8, adapted from <https://github.com/pydanny/cached-property>
  class cached_property(object):
    def __init__(self, fn):
      self.fn = fn
    def __get__(self, obj, cls):
      value = obj.__dict__[self.fn.__name__] = self.fn(obj)
      return value
logger = ir_datasets.log.easy()


class CorpusGraph(pta.Artifact):
  def neighbours(self, docid: Union[int, str], weights: bool = False) -> Union[np.array, List[str], Tuple[np.array, np.array], Tuple[List[str], np.array]]:
    raise NotImplementedError()

  @staticmethod
  def load(path, **kwargs):
    with (Path(path)/'pt_meta.json').open('rt') as fin:
      meta = json.load(fin)
    assert meta.get('type') == 'corpus_graph'
    fmt = meta.get('format')
    if fmt == 'np_topk' or fmt == 'numpy_kmax':
      return NpTopKCorpusGraph(path, **kwargs)
    raise ValueError(f'Unknown corpus graph format: {fmt}')

  # TODO: rework & verify
  # @staticmethod
  # def from_np_dense_index(np_index, out_dir: Union[str, Path], k: int = 16, batch_size: int = 50000) -> 'CorpusGraph':
  #   out_dir = Path(out_dir)
  #   if out_dir.exists():
  #     raise FileExistsError(out_dir)
  #   out_dir.mkdir(parents=True)
  #   edges_path = out_dir/'edges.u32.np'
  #   weights_path = out_dir/'weights.f16.np'
  #   shutil.copyfile(np_index.index_path/'docnos.np', out_dir/'docnos.np')
  #   S = batch_size
  #   docnos, vectors = np_index.docnos_and_data()
  #   doc_count = 0
  #   with ir_datasets.util.finialized_file(str(edges_path), 'wb') as fe, ir_datasets.util.finialized_file(str(weights_path), 'wb') as fw:
  #     for i in logger.pbar(range(0, vectors.shape[0], S)):
  #       top_idxs = []
  #       top_scores = []
  #       left = torch.from_numpy(vectors[i:i+S]).cuda()
  #       left /= left.norm(dim=1, keepdim=True)
  #       for j in logger.pbar(range(0, vectors.shape[0], S)):
  #         right = torch.from_numpy(vectors[j:j+S]).cuda()
  #         right /= right.norm(dim=1, keepdim=True)
  #         top = (left @ right.T).topk(k+1, sorted=True, dim=1)
  #         top_idxs.append((top.indices + j).cpu())
  #         top_scores.append(top.values.cpu())
  #       top_idxs = torch.cat(top_idxs, dim=1)
  #       top_scores = torch.cat(top_scores, dim=1).cuda()
  #       top = top_scores.topk(k+1, sorted=True, dim=1)
  #       idxs = torch.gather(top_idxs, 1, top.indices.cpu())[:, 1:]
  #       sims = top.values.cpu()[:, 1:]
  #       fe.write(idxs.cpu().numpy().astype(np.uint32).tobytes())
  #       fw.write(sims.numpy().astype(np.float16).tobytes())
  #       doc_count += left.shape[0]
  #   with (out_dir/'pt_meta.json').open('wt') as fout:
  #     json.dump({
  #       'type': 'corpus_graph',
  #       'format': 'np_topk',
  #       'doc_count': doc_count,
  #       'k': k,
  #     }, fout)
  #   return NumpyKMaxCorpusGraph(out_dir)

  @staticmethod
  def from_retriever(retriever, docs_it, out_dir: Union[str, Path], k: int = 16, batch_size: int = 1024) -> 'CorpusGraph':
    out_dir = Path(out_dir)
    if out_dir.exists():
      raise FileExistsError(out_dir)
    out_dir.mkdir(parents=True)
    edges_path = out_dir/'edges.u32.np'
    weights_path = out_dir/'weights.f16.np'

    # First step: We need a docno <-> index mapping for this to work. Do a pass over the iterator
    # to build a docno loookup file, while also writing the contents to a temporary file that we'll read
    # in the next loop. (This avoids loading the entire iterator into memory.)
    with tempfile.TemporaryDirectory() as dout:
      with LZ4FrameFile(f'{dout}/docs.pkl.lz4', 'wb') as fout, Lookup.builder(out_dir/'docnos.npids') as docno_builder:
        for doc in logger.pbar(docs_it, miniters=1, smoothing=0, desc='first pass'):
          pickle.dump(doc, fout)
          docno_builder.add(doc['docno'])
      docnos = Lookup(out_dir/'docnos.npids')

      # Now read out everything in the file and retrieve using each one. Do this in batches for efficiency.
      with ir_datasets.util.finialized_file(str(edges_path), 'wb') as fe, ir_datasets.util.finialized_file(str(weights_path), 'wb') as fw, LZ4FrameFile(f'{dout}/docs.pkl.lz4', 'rb') as fin:
        for chunk in more_itertools.chunked(logger.pbar(range(len(docnos)), miniters=1, smoothing=0, desc='searching', total=len(docnos)), batch_size):
          chunk = [pickle.load(fin) for _ in chunk]
          res = retriever(pd.DataFrame(chunk).rename(columns={'docno': 'qid', 'text': 'query'}))
          res_by_qid = dict(iter(res.groupby('qid')))
          for docno in [c['docno'] for c in chunk]:
            did_res = res_by_qid.get(docno)
            dids, scores = [], []
            if did_res is not None:
              did_res = did_res[did_res.docno != docno].iloc[:k]
              if len(did_res) > 0:
                dids = docnos.inv[list(did_res.docno)]
                scores = list(did_res.score)
            if len(dids) < k: # if we didn't get as many as we expect, loop the document back to itself.
              dids += [docnos.inv[docno]] * (k - len(dids))
              scores += [0.] * (k - len(scores))
            fe.write(np.array(dids, dtype=np.uint32).tobytes())
            fw.write(np.array(scores, dtype=np.float16).tobytes())

    # Finally, keep track of metadata about this artefact.
    with (out_dir/'pt_meta.json').open('wt') as fout:
      json.dump({
        'type': 'corpus_graph',
        'format': 'np_topk',
        'package_hint': 'pyterrier-adaptive',
        'doc_count': len(docnos),
        'k': k,
      }, fout)

    return NpTopKCorpusGraph(out_dir)


class NpTopKCorpusGraph(CorpusGraph):
  def __init__(self, path, k=None):
    super().__init__(path)
    with (self.path/'pt_meta.json').open('rt') as fin:
      self.meta = json.load(fin)
    assert self.meta.get('type') == 'corpus_graph' and self.meta.get('format') in ('numpy_kmax', 'np_topk')
    self._data_k = self.meta['k']
    if k is not None:
      assert k <= self.meta['k']
    self._k = self.meta['k'] if k is None else k
    self._edges_path = self.path/'edges.u32.np'
    self._weights_path = self.path/'weights.f16.np'
    self._docnos = Lookup(self.path/'docnos.npids')

  def __repr__(self):
    return f'NpTopKCorpusGraph({repr(str(self.path))}, k={self._k})'

  @cached_property
  def edges_data(self):
    res = np.memmap(self._edges_path, mode='r', dtype=np.uint32).reshape(-1, self._data_k)
    if self._k != self._data_k:
      res = res[:, :self._k]
    return res

  @cached_property
  def weights_data(self):
    res = np.memmap(self._weights_path, mode='r', dtype=np.float16).reshape(-1, self._data_k)
    if self._k != self._data_k:
      res = res[:, :self._k]
    return res

  def neighbours(self, docid, weights=False):
    as_str = isinstance(docid, str)
    if as_str:
      docid = self._docnos.inv[docid]
    neigh = self.edges_data[docid]
    if as_str:
      neigh = self._docnos.fwd[neigh]
    if weights:
      weigh = self.weights_data[docid]
      return neigh, weigh
    return neigh

  def to_limit_k(self, k: int) -> 'NpTopKCorpusGraph':
    """
    Creates a version of the graph with a maximum of k edges from each node.
    k must be less than the number of edges per node from the original graph.
    """
    return NpTopKCorpusGraph(self.path, k)



__version__ = '0.2.0'

from .gar import GAR
from .corpus_graph import CorpusGraph, NpTopKCorpusGraph
from pyterrier_adaptive._laff import Laff

__all__ = ['GAR', 'CorpusGraph', 'NpTopKCorpusGraph', 'Laff']



from typing import Optional
import numpy as np
from collections import Counter
import pyterrier as pt
import pandas as pd
import ir_datasets
import pyterrier_adaptive
logger = ir_datasets.log.easy()


class GAR(pt.Transformer):
    """A :class:`~pyterrier.Transformer` that implements Graph-based Adaptive Re-ranking (GAR).

    Required input columns: ``['qid', 'query', 'docno', 'score', 'rank']``

    Output columns: ``['qid', 'query', 'docno', 'score', 'rank', 'iteration']``

    .. note::

        The iteration column defines the batch number that first identified the document in the
        results. Due to the alternating nature of the algorithm, ``even=initial retrieval``, ``odd=corpus graph``,
        and ``-1=backfilled``.
    
    .. cite.dblp:: conf/cikm/MacAvaneyTM22
    """
    def __init__(self,
        scorer: pt.Transformer,
        corpus_graph: 'pyterrier_adaptive.CorpusGraph',
        num_results: int = 1000,
        batch_size: Optional[int] = None,
        backfill: bool = True,
        enabled: bool = True,
        verbose: bool = False):
        """
        Args:
            scorer(:class:`~pyterrier.Transformer`): A transformer that scores query-document pairs. It will only be provided with ['qid, 'query', 'docno', 'score'].
            corpus_graph(:class:`~pyterrier_adaptive.CorpusGraph`): A graph of the corpus, enabling quick lookups of nearest neighbours
            num_results(int): The maximum number of documents to score (called "budget" and $c$ in the paper)
            batch_size(int): The number of documents to score at once (called $b$ in the paper). If not provided, will attempt to use the batch size from the scorer
            backfill(bool): If True, always include all documents from the initial stage, even if they were not re-scored
            enabled(bool): If False, perform re-ranking without using the corpus graph
            verbose(bool): If True, print progress information
        """
        self.scorer = scorer
        self.corpus_graph = corpus_graph
        self.num_results = num_results
        if batch_size is None:
            batch_size = scorer.batch_size if hasattr(scorer, 'batch_size') else 16
        self.batch_size = batch_size
        self.backfill = backfill
        self.enabled = enabled
        self.verbose = verbose

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Applies Graph-based Adaptive Re-ranking to the provided dataframe. Essentially,
        Algorithm 1 from the paper.
        """
        result = {'qid': [], 'query': [], 'docno': [], 'rank': [], 'score': [], 'iteration': []}

        df = dict(iter(df.groupby(by=['qid'])))
        qids = df.keys()
        if self.verbose:
            qids = logger.pbar(qids, desc='adaptive re-ranking', unit='query')

        for qid in qids:
            query = df[qid]['query'].iloc[0]
            scores = {}
            res_map = [Counter(dict(zip(df[qid].docno, df[qid].score)))] # initial results
            if self.enabled:
                res_map.append(Counter()) # frontier
            frontier_data = {'minscore': float('inf')}
            iteration = 0
            while len(scores) < self.num_results and any(r for r in res_map):
                if len(res_map[iteration%len(res_map)]) == 0:
                    # if there's nothing available for the one we select, skip this iteration (i.e., move on to the next one)
                    iteration += 1
                    continue
                this_res = res_map[iteration%len(res_map)] # alternate between the initial ranking and frontier
                size = min(self.batch_size, self.num_results - len(scores)) # get either the batch size or remaining budget (whichever is smaller)
                
                # build batch of documents to score in this round
                batch = this_res.most_common(size)
                batch = pd.DataFrame(batch, columns=['docno', 'score'])
                batch['qid'] = qid
                batch['query'] = query

                # go score the batch of document with the re-ranker
                batch = self.scorer(batch)
                scores.update({k: (s, iteration) for k, s in zip(batch.docno, batch.score)})
                self._drop_docnos_from_counters(batch.docno, res_map)
                if len(scores) < self.num_results and self.enabled:
                    self._update_frontier(batch, res_map[1], frontier_data, scores)
                iteration += 1

            # Add scored items to results
            result['qid'].append(np.full(len(scores), qid))
            result['query'].append(np.full(len(scores), query))
            result['rank'].append(np.arange(len(scores)))
            for did, (score, i) in Counter(scores).most_common():
                result['docno'].append(did)
                result['score'].append(score)
                result['iteration'].append(i)

            # Backfill unscored items
            if self.backfill and len(scores) < self.num_results:
                last_score = result['score'][-1] if result['score'] else 0.
                count = min(self.num_results - len(scores), len(res_map[0]))
                result['qid'].append(np.full(count, qid))
                result['query'].append(np.full(count, query))
                result['rank'].append(np.arange(len(scores), len(scores) + count))
                for i, (did, score) in enumerate(res_map[0].most_common()):
                    if i >= count:
                        break
                    result['docno'].append(did)
                    result['score'].append(last_score - 1 - i)
                    result['iteration'].append(-1)

        return pd.DataFrame({
            'qid': np.concatenate(result['qid']),
            'query': np.concatenate(result['query']),
            'docno': result['docno'],
            'rank': np.concatenate(result['rank']),
            'score': result['score'],
            'iteration': result['iteration'],
        })

    def _update_frontier(self, scored_batch, frontier, frontier_data, scored_dids):
        remaining_budget = self.num_results - len(scored_dids)
        for score, did in sorted(zip(scored_batch.score, scored_batch.docno), reverse=True):
            if len(frontier) < remaining_budget or score >= frontier_data['minscore']:
                hit = False
                for target_did in self.corpus_graph.neighbours(did):
                    if target_did not in scored_dids:
                        if target_did not in frontier or score > frontier[target_did]:
                            frontier[target_did] = score
                            hit = True
                if hit and score < frontier_data['minscore']:
                    frontier_data['minscore'] = score

    def _drop_docnos_from_counters(self, docnos, counters):
        for docno in docnos:
            for c in counters:
                del c[docno]



from typing import Optional, Union, List
import shutil
import torch
import numpy as np
import pandas as pd
from tqdm import tqdm
from itertools import count
from more_itertools import chunked
import pyterrier as pt
import pyterrier_alpha as pta
from pyterrier_adaptive import CorpusGraph, NpTopKCorpusGraph


class Laff(pt.Transformer):
    """A transformer that computes a learned affinity score between two document texts using a transformer model.

    .. cite.dblp::     journals/corr/abs-2410-20286
    """

    def __init__(self,
        model: str = 'macavaney/laff',
        *,
        device: Optional[Union[str, torch.device]] = None,
        batch_size: int = 128,
        max_length: int = 512,
        verbose: bool = False
    ):
        """
        Args:
            model: the name of the transformer model to use.
            device: the device to use for the transformer model.
            batch_size: the batch size to use for processing.
            max_length: the maximum length of the input text.
            verbose: whether to display progress bars.
        """
        from transformers import AutoModelForSequenceClassification, AutoTokenizer
        self.model = AutoModelForSequenceClassification.from_pretrained(model)
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.device = torch.device(device or ('cuda' if torch.cuda.is_available() else 'cpu'))
        self.model = self.model.to(self.device)
        self.batch_size = batch_size
        self.max_length = max_length
        self.verbose = verbose

    def compute_affinity(self,
        texts_left: Union[str, List[str]],
        texts_right: Union[str, List[str]]
    ) -> List[float]:
        """ Compute the affinity scores between two lists of texts.

        Args:
            texts_left: the left-hand side texts.
            texts_right: the right-hand side texts.

        If either the left or right text is a string (or length-1 list), it is projected to the
        length of the other input (akin to numpy or torch projection).

        A higher affinity score indicates the documents are more similar to one another.

        .. code-block:: python
            :caption: Compute the Learned Affinity (LAFF) score between documents.

            >>> from pyterrier_adaptive import Laff
            >>> model = Laff()
            >>> model.compute_affinity('the cat sat on the mat', ['cats like to sit in the sun', 'dogs like to play fetch'])
            [5.46875, -3.140625]

        Returns:
            A list of affinity scores.
        """
        if isinstance(texts_left, str):
            texts_left = [texts_left]
        if isinstance(texts_right, str):
            texts_right = [texts_right]
        if len(texts_left) == 1:
            texts_left = texts_left * len(texts_right)
        elif len(texts_right) == 1:
            texts_right = texts_right * len(texts_left)
        assert len(texts_left) == len(texts_right)

        affinity_scores = []

        it = chunked(zip(texts_left, texts_right), self.batch_size)
        if self.verbose:
            it = tqdm(it, unit='batch', total=len(texts_left))

        for batch in it:
            batch = list(batch)
            batch_left, batch_right = zip(*batch)
            enc = self.tokenizer(
                list(batch_left),
                list(batch_right),
                max_length=self.max_length,
                padding='longest',
                truncation=True,
                return_attention_mask=True,
                return_tensors='pt',
            ).to(self.device)

            with torch.no_grad(), torch.autocast(device_type=self.device.type):
                outputs = self.model(**enc)
                affinity_scores += outputs.logits.flatten().tolist()

        return affinity_scores

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """ Compute the affinity scores between two columns of texts in the input.

        Expects columns ``text`` (for the left-side text) and ``other_text`` (for the right-side text).

        Results are sorted by the left-side text and the affinity score.
        """
        pta.validate.columns(inp, includes=['text', 'other_text'])
        affinity = self.compute_affinity(inp['text'], inp['other_text'])
        res = inp.assign(affinity=affinity)
        res.sort_values(['text', 'affinity'], ascending=[True, False], inplace=True)
        return res

    def wrap_graph(self,
        graph: CorpusGraph,
        text_loader: pt.Transformer,
    ) -> 'OnTheFlyLaffGraph':
        """Wrap a corpus graph with the LAFF transformer for on-the-fly LAFF scomre computation.

        Args:
            graph: the input corpus graph.
            text_loader: a transformer that loads the text for a given document.

        Returns:
            A corpus graph that computes LAFF scores on-the-fly.
        """
        return OnTheFlyLaffGraph(graph, self, text_loader)

    def apply_to_graph(self,
        graph: NpTopKCorpusGraph,
        text_loader: pt.Transformer,
        out_path: Optional[str] = None,
        *,
        verbose: Optional[bool] = None
    ) -> NpTopKCorpusGraph:
        """ Apply the LAFF transformer to a corpus graph to construct a new one.

        Args:
            graph: the input corpus graph.
            text_loader: a transformer that loads the text for a given document.
            out_path: the path to save the output corpus graph. If not provided, the input graph's path is used with a '.laff' extension.
            verbose: whether to display progress bars.
        """
        if out_path is None:
            out_path = str(graph.path) + '.laff'
        if verbose is None:
            verbose = self.verbose

        with pta.ArtifactBuilder(NpTopKCorpusGraph, path=out_path) as b:
            shutil.copyfile(graph.path / 'docnos.npids', b.path / 'docnos.npids')
            docnos = graph._docnos
            doc_count = b.metadata['doc_count'] = len(docnos)
            k = b.metadata['k'] = graph._k
            with (b.path/'edges.u32.np').open('wb') as fe, (b.path/'weights.f16.np').open('wb') as fw:
                it = zip(count(), docnos, graph.edges_data)
                if verbose:
                    it = tqdm(it, unit='doc', total=doc_count)
                for did, docno, neighbor_ids in it:
                    neighbor_ids = neighbor_ids[neighbor_ids != did] # ignore self-links
                    texts = text_loader(pd.DataFrame({'docno': [docno] + list(docnos.fwd[neighbor_ids])}))['text']
                    this_text, other_texts = texts[0], texts[1:]
                    affinity = np.array(self.compute_affinity(this_text, other_texts))
                    sort = np.argsort(-affinity)
                    weights = affinity[sort].astype(np.float16)
                    edges = neighbor_ids[sort].astype(np.uint32)
                    if len(weights) < k: # pad up to k with self-links if needed
                        weights = np.pad(weights, (0, k - len(weights)), constant_values=float('-inf'))
                        edges = np.pad(edges, (0, k - len(edges)), constant_values=did)
                    fe.write(edges.tobytes())
                    fw.write(weights.tobytes())

        return NpTopKCorpusGraph(out_path)


class OnTheFlyLaffGraph:
    def __init__(self, graph: CorpusGraph, laff: Laff, text_loader: pt.Transformer):
        self.graph = graph
        self.laff = laff
        self.text_loader = text_loader

    def neighbours(self, docid: str, weights: bool = False, orig_weights: bool = False):
        orig_neighbors, orig_weights_value = self.graph.neighbours(docid, weights=True)
        orig_count = len(orig_neighbors)
        orig_weights_value = [w for w, n in zip(orig_weights_value, orig_neighbors) if n != docid]
        orig_neighbors = [n for n in orig_neighbors if n != docid]
        left_text, *right_texts = self.text_loader(pd.DataFrame({'docno': [docid] + orig_neighbors}))['text']
        affinity_scores = self.laff.compute_affinity(left_text, right_texts)
        affinity_scores = np.array(affinity_scores)
        sort_order = (-affinity_scores).argsort()
        new_neighbors = [orig_neighbors[i] for i in sort_order]
        if len(new_neighbors) < orig_count:
            new_neighbors += [docid] * (orig_count - len(new_neighbors))
        result = [new_neighbors]
        if weights:
            affinity_scores = [affinity_scores[i] for i in sort_order]
            if len(affinity_scores) < orig_count:
                affinity_scores += [float('-inf')] * (orig_count - len(affinity_scores))
            result.append(affinity_scores)
        if orig_weights:
            orig_weights_value = [orig_weights_value[i] for i in sort_order]
            if len(orig_weights_value) < orig_count:
                orig_weights_value += [float('-inf')] * (orig_count - len(orig_weights_value))
            result.append(orig_weights_value)
        if len(result) == 1:
            return result[0]
        return tuple(result)
